//
// Created by Martí Rubio Estrada on 19/11/2019.
//

#include <cmath>
#include "Serie.h"


/**
 * Constructor de Sèrie.
 * Crea un polinomi de grau degree amb coeficients zero
 * @param degree grau del polinomi a crear
 */
inline Serie::Serie(int degree)
{
    this->coeff = new double[degree + 1];
    this->degree = degree;
    for(unsigned int i = 0; i <= degree; i++){
        this->coeff[i] = 0.0;
    }
}

/**
 * Constructor de Sèrie.
 * Crea un polinomi de grau DEGREE (constant global triada a main.cpp) amb coeficients zero
 */
inline Serie::Serie()
{
    this->coeff = new double[DEGREE + 1];
    this->degree = DEGREE;
    for(unsigned int i = 0; i <= DEGREE; i++){
        this->coeff[i] = 0.0;
    }
}

/**
 * Constructor de Sèrie.
 * Crea un polinomi de grau degree amb coeficients segons in_coeff
 * @param in_coeff valors dels coeficients
 * @param degree grau del polinomi a crear
 */
inline Serie::Serie(double* in_coeff, int degree)
{
    this->degree = degree;
    this->coeff = new double[degree + 1];

    for(unsigned int i = 0; i <= degree; i++){
        this->coeff[i] = in_coeff[i];
    }

}

/**
 * Destructor de Sèrie.
 */
inline Serie::~Serie(){
    if(this->coeff != nullptr){
        delete[] this->coeff;
    }
}

/**
 * Getter del grau del polinomi.
 * @return grau del polinomi
 */
inline int Serie::getDegree() const
{
    return this->degree;
}

/**
 * Getter dels coeficients.
 * @return punter a l'array de coeficients
 */
inline double* Serie::getCoeff() const
{
    return this->coeff;
}

/**
 * Getter del coeficient en posició position.
 * @param position grau del coeficient a retornar
 * @return coeficient
 */
inline double Serie::getCoeff(int position) const
{
    if(position > this->degree){
        return 0;
    }
    return this->coeff[position];
}

/**
 * Setter del coeficient en posició position.
 * @param value valor a settejar
 * @param position grau del coeficient a canviar
 */
inline void Serie::setCoeff(double value, int position){
    if(position<=this->degree){
        this->coeff[position] = value;
    }
}

/**
 * Suma un valor al coeficient de grau position.
 * @param value valor a sumar
 * @param position grau del valor a sumar
 */
inline void Serie::addCoeff(double value, int position){
    if(position<=this->degree){
        this->coeff[position] += value;
    }
}

/**
 * Sobrecàrrega de l'operador =
 * @param serie serie a posar a this
 * @return this
 */
inline Serie& Serie::operator=(const Serie serie)
{


    this->coeff = new double[serie.getDegree() + 1];


    for(int i = 0; i <= serie.getDegree(); i++){
        this->coeff[i] = serie.getCoeff(i);
    }

    this->degree = serie.getDegree();

    return *this;
}


/**
 * Sobrecàrrega de l'operador =
 * @param serie punter a la sèrie a posar a this
 * @return this
 */
inline Serie& Serie::operator=(const Serie* serie)
{

    this->coeff = new double[serie->getDegree() + 1];


    for(int i = 0; i <= serie->getDegree(); i++){
        this->coeff[i] = serie->getCoeff(i);
    }

    this->degree = serie->getDegree();

    return *this;
}

/**
 * Sobrecàrrega de l'operador +
 * @param serie polinomi a sumar amb this
 * @return polinomi sumat
 */
inline Serie Serie::operator+(const Serie serie)
{

    if (serie.getDegree() >= this->getDegree()){
        Serie aux(serie.getDegree());

        for(unsigned int i = 0; i <= this->getDegree(); i++){
            aux.setCoeff(this->coeff[i] + serie.getCoeff()[i], i);
        }

        for(unsigned int i = this->getDegree()+1; i <= serie.getDegree(); i++){
            aux.setCoeff(serie.getCoeff()[i], i);
        }

        return aux;
    }
    else{
        Serie aux(this->getDegree());

        for(unsigned int i = 0; i <= serie.getDegree(); i++){
            aux.setCoeff(this->coeff[i] + serie.getCoeff()[i], i);
        }


        for(unsigned int i = serie.getDegree()+1; i <= this->getDegree(); i++){
            aux.setCoeff(this->coeff[i], i);
        }
        //Serie aux(result, this->getDegree());
        //Serie* result_serie = new Serie(result, serie.getDegree());
        return aux;
    }
}

/**
 * Sobrecàrrega de l'operador +
 * @param serie punter al polinomi a sumar
 * @return polinomi sumat
 */
inline Serie Serie::operator+(const Serie* serie)
{


    if (serie->getDegree() >= this->getDegree()){
        Serie aux(serie->getDegree());

        for(unsigned int i = 0; i <= this->getDegree(); i++){
            aux.setCoeff(this->coeff[i] + serie->getCoeff()[i], i);
        }

        for(unsigned int i = this->getDegree()+1; i <= serie->getDegree(); i++){
            aux.setCoeff(serie->getCoeff()[i], i);
        }

        return aux;
    }
    else{
        Serie aux(this->getDegree());

        for(unsigned int i = 0; i <= serie->getDegree(); i++){
            aux.setCoeff(this->coeff[i] + serie->getCoeff()[i], i);
        }


        for(unsigned int i = serie->getDegree()+1; i <= this->getDegree(); i++){
            aux.setCoeff(this->coeff[i], i);
        }
        //Serie aux(result, this->getDegree());
        //Serie* result_serie = new Serie(result, serie.getDegree());
        return aux;
    }
}

/**
 * Sobrecàrrega de l'operador -
 * @param serie Serie a restar
 * @return polinomi restat
 */
inline Serie Serie::operator-(const Serie serie)
{
    if (serie.getDegree() >= this->getDegree()){
        Serie aux(serie.getDegree());

        for(unsigned int i = 0; i <= this->getDegree(); i++){
            aux.setCoeff(this->coeff[i] - serie.getCoeff()[i], i);
        }


        for(unsigned int i = this->getDegree()+1; i <= serie.getDegree(); i++){
            aux.setCoeff(- serie.getCoeff()[i], i);
        }

        //Serie aux(result, serie.getDegree());
        //Serie* result_serie = new Serie(result, serie.getDegree());
        return aux;
    }
    else{
        //double* result = new double[this->getDegree() + 1];
        Serie aux(this->getDegree());
        for(unsigned int i = 0; i <= serie.getDegree(); i++){
            aux.setCoeff(this->coeff[i] - serie.getCoeff()[i],i);
        }


        for(unsigned int i = serie.getDegree()+1; i <= this->getDegree(); i++){
            aux.setCoeff(this->getCoeff()[i], i);
        }

        //Serie aux(result, this->getDegree());
        //Serie* result_serie = new Serie(result, serie.getDegree());
        return aux;
    }

}

/**
 * Sobrecàrrega de l'operador *
 * @param serie Serie a multiplicar
 * @return polinomi multiplicat
 */
inline Serie Serie::operator*(const Serie serie)
{



    if(serie.getDegree() >= this->getDegree()){
        Serie aux(serie.getDegree());
        for(unsigned int i = 0; i <= serie.getDegree(); i++){
            aux.setCoeff(0,i);
        }

        for(unsigned int i = 0; i <= this->getDegree(); i++){
            for(unsigned int j = 0; j <= serie.getDegree(); j++){
                if(i + j <= serie.getDegree()){
                    aux.addCoeff(this->coeff[i]*serie.getCoeff()[j], i+j);

                }
            }
        }
        return aux;
    }else{
        Serie aux(this->getDegree());
        for(unsigned int i = 0; i <= this->getDegree(); i++){
            aux.setCoeff(0,i);
        }

        for(unsigned int i = 0; i <= this->getDegree(); i++){
            for(unsigned int j = 0; j <= serie.getDegree(); j++){
                if(i + j <= this->getDegree()){
                    aux.addCoeff(this->coeff[i]*serie.getCoeff()[j], i+j);

                }
            }
        }
        return aux;
    }

}

/**
 * Sobrecàrrega de l'operador /
 * @param serie Sèrie a dividir
 * @return polinomi dividit
 */
inline Serie Serie::operator/(const Serie serie)
{
    Serie aux(serie.getDegree());
    double* result = new double[serie.getDegree() + 1];
    for(unsigned int i = 0; i <= serie.getDegree(); i++){
        aux.setCoeff(0,i);
    }

    aux.setCoeff(this->coeff[0]/serie.getCoeff()[0], 0);

    for(unsigned int i = 1; i <= this->getDegree(); i++){
        double sum = 0;
        for(unsigned int j = 0; j <= i - 1; j++){
            sum += aux.getCoeff(j)*serie.getCoeff()[i - j];
        }
        aux.setCoeff((1.0/serie.getCoeff()[0])*(this->coeff[i] - sum), i);

    }
    return aux;

}

/**
 * Sobrecàrrega de l'operador +
 * @param value número a sumar
 * @return polinomi sumat
 */
inline Serie Serie::operator+(double value)
{
    Serie aux(this->degree);
    aux.setCoeff(value + this->getCoeff(0),0);
    for(int i = 1; i <= this->getDegree(); i++){
        aux.setCoeff(this->getCoeff()[i], i);
    }

    return aux;
}

/**
 * Sobrecàrrega de l'operador -
 * @param value Número a restar
 * @return polinomi restat
 */
inline Serie Serie::operator-(double value)
{
    Serie aux(this->degree);
    for(int i = 0; i <= this->getDegree(); i++){
        aux.setCoeff(this->getCoeff()[i], i);
    }
    aux.addCoeff(-value, 0);
    return aux;
}

/**
 * Sobrecàrrega de l'operador *
 * @param value Número a multiplicar
 * @return polinomi multiplicat
 */
inline Serie Serie::operator*(double value)
{
    Serie aux(this->degree);
    for(int i = 0; i <= this->getDegree(); i++){
        aux.setCoeff(value*this->getCoeff(i), i);
    }
    return aux;
}

/**
 * Sobrecàrrega de l'operador /
 * @param value Número a dividir
 * @return polinomi dividit
 */
inline Serie Serie::operator/(double value)
{
    Serie aux(this->degree);
    for(int i = 0; i <= this->getDegree(); i++){
        aux.setCoeff(this->getCoeff()[i]/value,i);
    }
    return aux;
}

/**
 * Sobrecàrrega de l'operador +
 * @param value Número a sumar
 * @param serie Serie a sumar
 * @return polinomi sumat
 */
inline Serie operator+(double value, Serie serie)
{
    Serie aux(serie.getDegree());
    for(int i = 0; i <= serie.getDegree(); i++){
        aux.setCoeff(serie.getCoeff()[i],i);
    }
    aux.addCoeff(value, 0);
    return aux;
}

/**
 * Sobrecàrrega de l'operador -
 * @param value Número a restar
 * @param serie Serie a restar
 * @return polinomi restat
 */
inline Serie operator-(double value, Serie serie)
{
    Serie aux(serie.getDegree());
    aux.setCoeff(value - serie.getCoeff()[0], 0);
    for(int i = 1; i <= serie.getDegree(); i++){
        aux.setCoeff(-serie.getCoeff()[i],i);
    }
    return aux;
}

/**
 * Sobrecàrrega de l'operador -
 * @return polinomi canviat de signe
 */
inline Serie Serie::operator-()
{
    Serie aux(this->degree);
    for(int i = 0; i <= this->getDegree(); i++){
        aux.setCoeff(-this->getCoeff()[i],i);
    }
    return aux;
}

/**
 * Sobrecàrrega de l'operador *
 * @param value valor a multiplicar
 * @param serie sèrie a multiplicar
 * @return polinomi multiplicat
 */
inline Serie operator*(double value, Serie serie)
{
    Serie aux(serie.getDegree());
    for(int i = 0; i <= serie.getDegree(); i++){
        aux.setCoeff(value*serie.getCoeff()[i],i);
    }
    return aux;
}

/**
 * Sobrecàrrega de l'operador +=
 * @param serie serie a sumar
 * @return this
 */
inline Serie Serie::operator+=(Serie serie)
{
    *this = serie + *this;
    return *this;
}

/**
 * Sobrecàrrega de l'operador -=
 * @param serie sèrie a restar
 * @return this
 */
inline Serie Serie::operator-=(Serie serie)
{
    *this = *this - serie;
    return *this;
}

/**
 * Sobrecàrrega de l'operador *=
 * @param serie sèrie a multiplicar
 * @return this
 */
inline Serie Serie::operator*=(const Serie serie)
{
    *this = *this * serie;
    return *this;
}

/**
 * Sobrecàrrega de l'operador -=
 * @param serie sèrie a restar
 * @return this
 */
inline Serie Serie::operator/=(const Serie serie)
{
    *this = *this / serie;
    return *this;
}

/**
 * Sobrecàrrega de l'operador +=
 * @param value número a restar
 * @return this
 */
inline Serie Serie::operator+=(double value)
{
    *this = value + *this;
    return *this;
}

/**
 * Sobrecàrrega de l'operador -=
 * @param value número a restar
 * @return this
 */
inline Serie Serie::operator-=(double value)
{
    *this = *this - value;
    return *this;
}

/**
 * Sobrecàrrega de l'operador *=
 * @param value número a multiplicar
 * @return this
 */
inline Serie Serie::operator*=(double value)
{
    *this = value * *this;
    return *this;
}

/**
 * Sobrecàrrega de l'operador /=
 * @param value número a restar
 * @return this
 */
inline Serie Serie::operator/=(double value)
{
    *this = *this / value;
    return *this;
}

/**
 * Sobrecàrrega de l'operador ==
 * @param serie sèrie a comparar
 * @return booleà si els dos polinomis són iguals
 */
inline bool Serie::operator==(const Serie serie)
{
    double* s1 = serie.getCoeff();

    for(unsigned int i = 0; i <= serie.getDegree(); i++){
        if(this->coeff[i] != s1[i]){
            return false;
        }
    }
    return true;
}

/**
 * Sobrecàrrega de l'operador !=
 * @param serie sèrie a comparar
 * @return booleà si els dos polinomis són diferents
 */
inline bool Serie::operator!=(const Serie serie)
{
    for(unsigned int i = 0; i <= serie.getDegree(); i++){
        if(this->coeff[i] != serie.getCoeff()[i]){
            return true;
        }
    }
    return false;
}

/**
 * Sobrecàrrega de l'operador ==
 * @param value número a comparar
 * @return booleà si el polinomi i el valor són iguals
 */
inline bool Serie::operator==(double value)
{
    for (unsigned int i = 0; i <= this->getDegree(); i++){
        if(i == 0){
            if(value != this->coeff[0]){
                return false;
            }
        }
        else{
            if(this->coeff[i] != 0.0){
                return false;
            }
        }
    }
    return true;
}

/**
 * Sobrecàrrega de l'operador !=
 * @param value número a comparar
 * @return booleà si el polinomi i el valor són diferents
 */
inline bool Serie::operator!=(double value)
{
    for (unsigned int i = 1; i <= this->getDegree(); i++){
        if(i == 0){
            if(value != this->coeff[0]){
                return true;
            }
        }
        else{
            if(this->coeff[i] != 0.0){
                return true;
            }
        }
    }
    return false;
}

/**
 * Sobrecàrrega de l'operador ==
 * @param value número a comparar
 * @param serie sèrie a comparar
 * @return booleà si el polinomi i el valor són iguals
 */
inline bool operator==(double value, Serie serie)
{
    for (unsigned int i = 0; i <= serie.getDegree(); i++){
        if(i == 0){
            if(value != serie.getCoeff()[0]){
                return false;
            }
        }
        else{
            if(serie.getCoeff()[i] != 0.0){
                return false;
            }
        }
    }
    return true;
}

/**
 * Sobrecàrrega de l'operador !=
 * @param value número a comparar
 * @return booleà si el polinomi i el valor són diferents
 */
inline bool operator!=(double value, Serie serie)
{
    for (unsigned int i = 0; i <= serie.getDegree(); i++){
        if(i == 0){
            if(value != serie.getCoeff()[0]){
                return true;
            }
        }
        else{
            if(serie.getCoeff()[i] != 0.0){
                return true;
            }
        }
    }
    return false;
}

/**
 * Funció de derivació de polinomi
 * @return polinomi derivat
 */
inline Serie Serie::derivate()
{
    Serie aux(this->degree);
    for (unsigned int i = 0; i < this->getDegree(); i++){
        aux.setCoeff((i + 1.0) * this->coeff[i+1],i);
    }
    aux.setCoeff(0,this->degree);
    return aux;
}

/**
 * Funció d'arrel d'un polinomi
 * @return arrel del polinomi
 */
inline Serie Serie::sqrt()
{
    Serie aux(this->degree);
    //double* result = new double[this->getDegree() + 1];
    for(unsigned int i = 0; i <= this->getDegree(); i++){
        aux.setCoeff(0,i);
    }

    aux.setCoeff(std::sqrt(this->coeff[0]),0);
    for(unsigned int i = 1; i <= this->getDegree(); i++){
        double sum = 0;
        for(unsigned int j = 1; j <= i - 1; j++){
            sum += aux.getCoeff(j)*aux.getCoeff(i-j);
        }
        aux.setCoeff((1.0/(2.0*aux.getCoeff(0)))*(this->coeff[i] - sum), i);
    }
    return aux;
}

/**
 * Funció de logaritme d'un polinomi
 * @return logaritme del polinomi
 */
inline Serie Serie::log()
{
    Serie derivate = this->derivate();
    Serie result = derivate/ *this;
    return result;
}

/**
 * Funció d'exponenciar un polinomi
 * @return exponencial del polinomi
 */
inline Serie Serie::exp()
{
    Serie aux(this->degree);
    for(unsigned int i = 0; i <= this->getDegree(); i++){
        aux.setCoeff(0,i);
    }

    aux.setCoeff(std::exp(this->coeff[0]), 0);

    for(unsigned int i = 1; i <= this->getDegree(); i++){
        double sum = 0;
        for(unsigned int j = 0; j <= i - 1; j++){
            sum += this->coeff[j]*aux.getCoeff(i - j - 1);
        }
        aux.setCoeff((1.0/i)*(sum), i);
    }
    return aux;

}

/**
 * Funció d'elevar un polinomi
 * @param a número al que elevem el polinomi
 * @return elevació del polinomi
 */
inline Serie Serie::pow(double a)
{
    Serie result = (this->log() * a).exp();
    return result;
}

/**
 * Funció de sinus d'un polinomi
 * @return sinus del polinomi
 */
inline Serie Serie::sin()
{
    Serie aux(this->degree);
    int deg = this->getDegree();

    double* a = new double[deg + 1];
    a[0] = std::cos(this->coeff[0]);
    aux.setCoeff(std::sin(this->coeff[0]),0);
    double a_sum;
    double b_sum;
    for(int n = 1; n <= deg; n++){
        a_sum = 0;
        b_sum = 0;
        for(int j = 1; j <= n; j++){
            a_sum += j*aux.getCoeff(n-j)*this->coeff[j];
            b_sum += j*a[n - j]*this->coeff[j];
        }
        a[n] = -(1.0/n)*a_sum;
        aux.setCoeff((1.0/n)*b_sum, n);
    }
    delete[] a;
    return aux;
}

/**
 * Funció de cosinus d'un polinomi
 * @return cosinus del polinomi
 */
inline Serie Serie::cos()
{
    Serie aux(this->degree);
    int deg = this->getDegree();
    double* b = new double[deg+1];
    aux.setCoeff(std::cos(this->coeff[0]),0);
    b[0] = std::sin(this->coeff[0]);
    double a_sum;
    double b_sum;
    for(int i = 1; i <= deg; i++){
        a_sum = 0;
        b_sum = 0;
        for(int j = 1; j <= i; j++){
            a_sum += j*b[i - j]*this->coeff[j];
            b_sum += j*aux.getCoeff(i-j)*this->coeff[j];
        }
        aux.setCoeff(-(1.0/i)*a_sum,i);
        b[i] = (1.0/i)*b_sum;
    }
    delete[] b;
    return aux;
}

/**
 * Avaluació del polinomi en un punt
 * @param s punt a avaluar
 * @return valor d'avaluació
 */
inline double Serie::calculate(double s)
{
    double result = 0.0;
    for(int i = 0; i <= this->getDegree(); i++){
        result += this->getCoeff(i)*::pow(s, i);
    }
    return result;
}

