//
// Created by Martí Rubio Estrada on 19/11/2019.
//
#import <cmath>
#include "differential_equation.h"
#include "Serie.cpp"

/**
 * Classe cercle que hereta de la classe abstracta differential_equation.
 * Equació diferencial que defineix un cercle a R2
 */
class Circle: public differential_equation
{
public:

    /**
     * Equació diferencial.
     * @param t temps d'avaluació
     * @param x punt d'avaluació
     * @param n dimensió del punt
     * @param result diferencial resultat
     */
    static void edo(double t, double x[], int n, double result[]){
        result[0] = -x[1];
        result[1] = x[0];
    }
};

/**
 * Classe cercle que hereta de la classe abstracta differential_equation_serie.
 * Equació diferencial que defineix un cercle a R2 treballant amb objectes sèries
 */
class Circle_Serie: public differential_equation_serie
{
public:

    /**
     * Equació diferencial.
     * @param t temps d'avaluació
     * @param x polinomi d'avaluació
     * @param n dimensió del punt
     * @param result polinomi resultat
     */
    static void edo(double t, Serie x[], int n, Serie **result){
        *result[0] = -x[1];
        *result[1] = x[0];
    }
};

/**
 * Classe pèndol pertorbat que hereta de la classe abstracta differential_equation.
 * Equació diferencial que defineix un pèndol pertorbat per un camp magnètic constant
 */
class Disturbed_Pendulum: public differential_equation
{
public:

    /**
     * Equació diferencial.
     * @param t temps d'avaluació
     * @param x punt d'avaluació
     * @param n dimensió del punt
     * @param result diferencial resultat
     */
    static void edo(double t, double x[], int n, double result[]){
        double epsilon = 0.1;
        double omega = sqrt(20);
        result[0] = x[1];
        result[1] = -sin(x[0]) + epsilon*sin(omega*t);
    }
};

/**
 * Classe matriu pèndol pertorbat que hereta de la classe abstracta differential_equation.
 * Equació diferencial i matriu diferencial que
 * defineix un pèndol pertorbat per un camp magnètic constant
 */
class Disturbed_Pendulum_Matrix: public differential_equation
{
public:

    /**
     * Equació diferencial.
     * @param t temps d'avaluació
     * @param x punt d'avaluació
     * @param n dimensió del punt
     * @param result diferencial resultat
     */
    static void edo(double t, double x[], int n, double result[]){
        double epsilon = 0.1;
        double omega = sqrt(20);
        result[0] = x[1];
        result[1] = -sin(x[0]) + epsilon*sin(omega*t);
        result[2] = x[4];
        result[3] = x[5];
        result[4] = -cos(x[0])*x[2];
        result[5] = -cos(x[0])*x[3];
    }
};

/**
 * Classe pèndol pertorbat que hereta de la classe abstracta differential_equation_serie
 * Equació diferencial que defineix un pèndol pertorbat per un
 * camp magnètic constant treballant amb objectes sèrie
 */
class Disturbed_Pendulum_Serie: public differential_equation_serie
{
public:

    /**
     * Equació diferencial.
     * @param t temps d'avaluació
     * @param x polinomi d'avaluació
     * @param n dimensió del punt
     * @param result polinomi resultat
     */
    static void edo(double t, Serie **x, int n, Serie **result){
        double epsilon = 0.1;
        double omega = sqrt(20);
        *result[0] = x[1];
        *result[1] = (-(x[0]->sin()) + epsilon*sin(omega*t));
    }
};