//
// Created by Martí Rubio Estrada on 06/01/2020.
//

#include <iostream>
#include <fstream>
#include <cmath>
#include "utils.h"
#include "rkv78.h"
#include "rkv78_serie.h"
#include "Serie.cpp"

using namespace std;


namespace utils{
    /**
     * variables auxiliars per al càlcul de les funcions, compartides entre elles.
     */
    double desc, module;
    double l1, l2;
    double v1[2], v2[2];
    double result[2], initial[2];
    double z1, z2;
    double initial_time, ah_copy;

    /**
     * Funció que a partir d'una bona aproximació troba un punt fix amb el mètode de Newton
     * @param at temps inicial
     * @param x punt inicial
     * @param n dimensió punt
     * @param ah pas d'integració
     * @param sc ús (1)/ no ús (0) del control de pas
     * @param tol tolerància màxima si usem control de pas
     * @param atf temps final d'integració
     * @param aer aturada (1)/ no aturada (0) del programa si es supera la tolerància
     * @param dif punter a l'equació a avaluar
     */
    void UtilFunctions::buscar_punt_fix(double *at, double *x, int n, double *ah, int sc, double tol,
              double *atf, double *aer, void (*dif)(double,double*,int,double*)){

        cout << "-------------- Entrem a la funció de buscar_punt_fix ---------------" << endl;
        ah_copy = *ah;
        initial_time = *at;
        result[0] = 1.0;
        result[1] = 1.0;

        double a1, b1, a2, b2, c1, c2;

        while(sqrt(result[0]*result[0] + result[1]*result[1]) > tol){

            // Posem la matriu a la identitat
            x[2] = 1.0;
            x[3] = 0.0;
            x[4] = 0.0;
            x[5] = 1.0;

            // Posem el pas inicial
            *at = initial_time;

            initial[0] = x[0];
            initial[1] = x[1];

            //Passos de l'algoritme Runge-Kutta
            while (*at < *atf){
                rkv78::rkv78::rkv78_main(at, x, n, ah, sc, tol, atf, aer, dif);
            }

            a1 = x[2] - 1.0;
            b1 = x[3];
            a2 = x[4];
            b2 = x[5] - 1.0;
            c1 = initial[0] - x[0];
            c2 = initial[1] - x[1];

            result[0] = (c1*b2 - c2*b1)/(a1*b2 - a2*b1);
            result[1] = (c2*a1 - c1*a2)/(a1*b2 - a2*b1);

            x[0] = initial[0] + result[0];
            x[1] = initial[1] + result[1];


            cout << "Error del punt actual: " << std::sqrt(result[0]*result[0] + result[1]*result[1]) << endl;
        }

        cout << "Error menor al demanat, comprovant el punt trobat..." << endl;

        cout << endl << "Comencem pel punt: " << x[0] << ", " << x[1] << endl;

        *at = initial_time;
        *ah = ah_copy;

        while (*at < *atf){
            rkv78::rkv78::rkv78_main(at, x, n, ah, 0, 1.e-8, atf, aer, dif);
        }

        cout << endl << "Acabem al punt: " << x[0] << ", " << x[1] << endl << endl;

    }

    /**
     * Funció que avalua l'algoritme de Runge-Kutta amb els valors demanats.
     * Aquesta funció s'ha de cridar si volem que el temps final sigui posterior al temps inicial.
     * @param at temps inicial
     * @param x punt inicial
     * @param n dimensió punt
     * @param ah pas d'integració (ha de ser positiu!)
     * @param sc ús (1)/ no ús (0) del control de pas
     * @param tol tolerància màxima si usem control de pas
     * @param atf temps final d'integració
     * @param aer aturada (1)/ no aturada (0) del programa si es supera la tolerància
     * @param dif punter a l'equació a avaluar
     */
    void UtilFunctions::evaluate_positive_time(double *at, double *x, int n, double *ah, int sc, double tol,
                                        double *atf, double *aer, void (*dif)(double,double*,int,double*)){

        cout << "-------------- Entrem a la funció evaluate amb temps positiu ---------------" << endl;
        initial_time = *at;
        while (*at < *atf){
            rkv78::rkv78::rkv78_main(at, x, n, ah, sc, tol, atf, aer, dif);
        }
    }

    /**
     * Funció que avalua l'algoritme de Runge-Kutta amb els valors demanats.
     * Aquesta funció s'ha de cridar si volem que el temps final sigui anterior al temps inicial.
     * @param at temps inicial
     * @param x punt inicial
     * @param n dimensió punt
     * @param ah pas d'integració (ha de ser negatiu!)
     * @param sc ús (1)/ no ús (0) del control de pas
     * @param tol tolerància màxima si usem control de pas
     * @param atf temps final d'integració
     * @param aer aturada (1)/ no aturada (0) del programa si es supera la tolerància
     * @param dif punter a l'equació a avaluar
     */
    void UtilFunctions::evaluate_negative_time(double *at, double *x, int n, double *ah, int sc, double tol,
                                 double *atf, double *aer, void (*dif)(double,double*,int,double*)){

        cout << "-------------- Entrem a la funció evaluate amb temps negatiu ---------------" << endl;
        initial_time = *at;
        while (*at > *atf){
            rkv78::rkv78::rkv78_main(at, x, n, ah, sc, tol, atf, aer, dif);
        }
    }

    /**
     * Funció que a partir d'un punt fix troba els valors i vectors propis usant el mètode de Cramer.
     * Aquest algoritme troba els valors propis només si són reals!
     * @param at temps inicial
     * @param x punt inicial
     * @param n dimensió punt
     * @param ah pas d'integració
     * @param sc ús (1)/ no ús (0) del control de pas
     * @param tol tolerància màxima si usem control de pas
     * @param atf temps final d'integració
     * @param aer aturada (1)/ no aturada (0) del programa si es supera la tolerància
     * @param dif punter a l'equació a avaluar
     */
    void UtilFunctions::buscar_vap_vep(double *at, double *x, int n, double *ah, int sc, double tol,
                        double *atf, double *aer, void (*dif)(double,double*,int,double*)) {

        cout << "-------------- Entrem a la funció buscar_vap_vep ---------------" << endl;
        double eq_a, eq_b, eq_c;
        initial_time = *at;
        x[2] = 1.0;
        x[3] = 0.0;
        x[4] = 0.0;
        x[5] = 1.0;
        *at = initial_time;

        while (*at < *atf){
            rkv78::rkv78::rkv78_main(at, x, n, ah, sc, tol, atf, aer, dif);
        }

        eq_a = 1.0;
        eq_b = x[2] + x[5];
        eq_c = x[2]*x[5] - x[3]*x[4];
        desc = eq_b*eq_b - 4.0*eq_a*eq_c;

        if (desc >= 0){
            l1 = (eq_b + sqrt(desc))/2.0;
            l2 = (eq_b - sqrt(desc))/2.0;

            v1[0] = 1.0;
            v1[1] = -(x[2] - l1)/x[3];

            module = sqrt(v1[0]*v1[0] + v1[1]*v1[1]);

            v1[0] = v1[0]/module;
            v1[1] = v1[1]/module;

            v2[0] = 1.0;
            v2[1] = -(x[2] - l2)/x[3];

            module = sqrt(v2[0]*v2[0] + v2[1]*v2[1]);

            v2[0] = v2[0]/module;
            v2[1] = v2[1]/module;

            cout << "vaps: " << l1 << ", " << l2 << endl;
            cout << "vep1: (" << v1[0] << ", " << v1[1] << ")" << endl;
            cout << "vep2: (" << v2[0] << ", " << v2[1] << ")" << endl << endl;
        }
    }

    /**
     * Funció que a partir d'un punt fix troba els valors i vectors propis usant el mètode de Cramer
     * @param at temps inicial
     * @param x punt inicial
     * @param n dimensió punt
     * @param ah pas d'integració
     * @param sc ús (1)/ no ús (0) del control de pas
     * @param tol tolerància màxima si usem control de pas
     * @param atf temps final d'integració
     * @param aer aturada (1)/ no aturada (0) del programa si es supera la tolerància
     * @param dif punter a l'equació a avaluar
     * @param vap1 punter al que s'introdueix el vap1
     * @param vap2 punter al que s'introdueix el vap2
     * @param vep1 punter al que s'introdueix el vep1
     * @param vep2 punter al que s'introdueix el vep2
     */
    void UtilFunctions::buscar_vap_vep(double *at, double *x, int n, double *ah, int sc, double tol,
                        double *atf, double *aer, void (*dif)(double,double*,int,double*),
                        double *vap1, double *vap2, double vep1[],double vep2[]){

        buscar_vap_vep(at, x, n, ah, sc, tol, atf, aer, dif);

        if(desc >= 0){
            *vap1 = l1;
            *vap2 = l2;

            for(int j = 0; j < 2; j++){
                vep1[j] = v1[j];
                vep2[j] = v2[j];
            }
        }
    }

    /**
     * Funció que busca la parametrització de les varietats estable o inestable d'un punt fix
     * @param at temps inicial
     * @param x punt inicial
     * @param n dimensió punt
     * @param ah pas d'integració
     * @param sc ús (1)/ no ús (0) del control de pas
     * @param tol tolerància màxima si usem control de pas
     * @param atf temps final d'integració
     * @param aer aturada (1)/ no aturada (0) del programa si es supera la tolerància
     * @param dif punter a l'equació a avaluar
     * @param deg grau màxim del polinomi
     * @param vap_a_usar flag que indica si trobem la varietat estable (0) / inestable (1)
     * @param dif_serie punter a l'equació a avaluar amb sèries
     */
    void UtilFunctions::buscar_varietat(double *at, double *x, int n, double *ah, int sc, double tol,
                         double *atf, double *aer, void (*dif)(double, double *, int, double *),
                         int deg, int vap_a_usar, void (*dif_serie)(double,Serie**,int,Serie**)) {

        cout << "-------------- Entrem a la funció buscar_varietat ---------------" << endl;
        initial_time = *at;

        // Crida a la funcio de buscar punt fix
        buscar_punt_fix(at, x, 6, ah, sc, tol, atf, aer, dif);

        *at = initial_time;

        // Crida a la funcio de buscar vaps i veps
        buscar_vap_vep(at, x, 6, ah, sc, tol, atf, aer, dif);

        z1 = x[0];
        z2 = x[1];

        int solved_degree = 1;
        double vap_used;
        double real_coef_x[deg + 1];
        double real_coef_y[deg + 1];

        real_coef_x[0] = z1;
        real_coef_y[0] = z2;

        double prova_x[1] = {z1};
        double prova_y[1] = {z2};

        Serie *p1 = new Serie(prova_x, 0);
        Serie *p2 = new Serie(prova_y, 0);

        if (vap_a_usar == 0) {
            // varietat estable
            if (l1 <= 1.0) {
                vap_used = l1;
            } else {
                vap_used = l2;
            }
        } else {
            // varietat inestable
            if (l1 <= 1.0) {
                vap_used = l2;
            } else {
                vap_used = l1;
            }
        }

        cout << "vap used: " << vap_used << endl << endl;

        if (vap_used == l1) {
            real_coef_x[1] = v1[0];
            real_coef_y[1] = v1[1];
        }
        else {
            real_coef_x[1] = v2[0];
            real_coef_y[1] = v2[1];
        }

        while (solved_degree < deg) {
            cout << "------------- deg: " << solved_degree + 1 << "---------------" << endl;

            double coeffs_x[solved_degree + 2];
            double coeffs_y[solved_degree + 2];

            for (int i = 0; i <= solved_degree; i++) {
                coeffs_x[i] = real_coef_x[i];
                coeffs_y[i] = real_coef_y[i];
            }

            coeffs_x[solved_degree + 1] = 0.0;
            coeffs_y[solved_degree + 1] = 0.0;

            Serie *s1 = new Serie(coeffs_x, solved_degree + 1);
            Serie *s2 = new Serie(coeffs_y, solved_degree + 1);

            Serie *pos[2];
            pos[0] = s1;
            pos[1] = s2;

            *at = initial_time;

            cout << "Polinomi x abans de començar:";
            for (int k = -1; k <= solved_degree; k++) {
                cout << " " << pos[0]->getCoeff()[k + 1]; //pow(vap_used, k+1);
            }
            cout << endl;

            cout << "Polinomi y abans de començar:";
            for (int k = -1; k <= solved_degree; k++) {
                cout << " " << pos[1]->getCoeff()[k + 1]; //pow(vap_used, k+1);
            }
            cout << endl;

            while (*at < *atf) {
                rkv78_serie::rkv78_serie::rkv78_main(at, pos, 2, ah, 0, 1e-12, atf, aer, dif_serie);
            }

            cout << "Polinomi x al final:";
            for (int k = -1; k <= solved_degree; k++) {
                cout << " " << pos[0]->getCoeff()[k + 1] / pow(vap_used, k + 1);
            }
            cout << endl;
            cout << "Polinomi y al final:";
            for (int k = -1; k <= solved_degree; k++) {
                cout << " " << pos[1]->getCoeff()[k + 1] / pow(vap_used, k + 1);
            }
            cout << endl << endl;

            double b_x = pos[0]->getCoeff()[solved_degree + 1];
            double b_y = pos[1]->getCoeff()[solved_degree + 1];
            double det = 1.0 / ((x[5] - pow(vap_used, solved_degree + 1)) * (x[2] - pow(vap_used, solved_degree + 1)) -
                                x[3] * x[4]);

            real_coef_x[solved_degree + 1] = det * ((pow(vap_used, solved_degree + 1) - x[5]) * b_x + x[3] * b_y);
            real_coef_y[solved_degree + 1] = det * (x[4] * b_x + (pow(vap_used, solved_degree + 1) - x[2]) * b_y);

            solved_degree++;
        }

        cout << endl << "Polinomi x definitiu:";
        for (int i = 0; i < deg + 1; i++) {
            cout << real_coef_x[i] << "  ";
        }
        cout << endl;
        cout << endl << "Polinomi y definitiu:";
        for (int i = 0; i < deg + 1; i++) {
            cout << real_coef_y[i] << "  ";
        }
    }
}